@@ -1,23 +1,32 @@
+import typing
 from os import environ
 
 
+class Constants(object):
+    default_webapi_address = "http://v3io-webapi:8081"
+
+
 class NuclioTrigger:
     kind = ""
 
-    def __init__(self, struct={}):
-        self._struct = struct
+    def __init__(self, struct: typing.Optional[dict] = None):
+        self._struct = struct or {}
 
     def to_dict(self):
         return self._struct
 
-    def disable(self, disabled=True):
+    def disable(self, disabled: bool = True):
         self._struct["disabled"] = disabled
         return self
 
-    def workers(self, workers=4):
-        self._struct["maxWorkers"] = workers
+    def workers(self, workers: int = 4):
+        self._struct["maxWorkers"] = int(workers)
         return self
 
+    @property
+    def get_workers(self) -> int:
+        return self._struct["maxWorkers"]
+
 
 class HttpTrigger(NuclioTrigger):
     kind = "http"
@@ -31,13 +40,17 @@ class HttpTrigger(NuclioTrigger):
         canary=None,
         secret=None,
     ):
-        self._struct = {
+        super(HttpTrigger, self).__init__({
             "kind": self.kind,
-            "maxWorkers": workers,
+            "maxWorkers": int(workers),
             "attributes": {"ingresses": {}},
             "annotations": {},
-        }
+        })
         if port:
+
+            # if port is a string for some reason, cast it
+            if not isinstance(port, int):
+                port = int(port)
             self._struct["attributes"]["port"] = port
         if host:
             self._ingress(host, paths, canary, secret=secret)
@@ -47,6 +60,14 @@ class HttpTrigger(NuclioTrigger):
     ):
         return self._ingress(host, paths, canary, name, secret)
 
+    @property
+    def get_port(self) -> typing.Optional[int]:
+        return self._struct["attributes"].get("port")
+
+    @property
+    def get_ingresses(self) -> typing.Optional[typing.Dict[str, dict]]:
+        return self._struct["attributes"]["ingresses"]
+
     def _ingress(
         self, host, paths=None, canary=None, name="0", secret=None,
     ):
@@ -83,11 +104,12 @@ class HttpTrigger(NuclioTrigger):
 class CronTrigger(NuclioTrigger):
     kind = "cron"
 
-    def __init__(self, interval="", schedule="", body="", headers={}):
-        self._struct = {
+    def __init__(self, interval="", schedule="", body="", headers=None):
+        super(CronTrigger, self).__init__({
             "kind": self.kind,
             "attributes": {},
-        }
+        })
+        headers = headers or {}
         if interval:
             self._struct["attributes"]["interval"] = interval
         elif schedule:
@@ -104,12 +126,13 @@ class CronTrigger(NuclioTrigger):
 class KafkaTrigger(NuclioTrigger):
     kind = "kafka"
 
-    def __init__(self, url, topic, partitions=[]):
-        self._struct = {
+    def __init__(self, url, topic, partitions=None):
+        super(KafkaTrigger, self).__init__({
             "kind": self.kind,
             "url": url,
             "attributes": {"topic": topic},
-        }
+        })
+        partitions = partitions or []
         if partitions:
             self._struct["attributes"]["partitions"] = partitions
 
@@ -139,20 +162,19 @@ class V3IOStreamTrigger(NuclioTrigger):
         container: str = None,
         path: str = None,
         workerAllocationMode: str = "pool",
-        webapi: str = "http://v3io-webapi:8081",
+        webapi: str = Constants.default_webapi_address,
         consumerGroup: str = "default",
         sequenceNumCommitInterval: str = "1s",
         heartbeatInterval: str = "3s",
     ):
-
         if url and not container and not path:
-            self._struct = {
+            struct = {
                 "kind": self.kind,
                 "url": url,
                 "attributes": {},
             }
         else:
-            self._struct = {
+            struct = {
                 "kind": self.kind,
                 "url": webapi,
                 "attributes": {
@@ -167,17 +189,27 @@ class V3IOStreamTrigger(NuclioTrigger):
             }
 
             if name:
-                self._struct["name"] = name
+                struct["name"] = name
 
         if maxWorkers:
-            self._struct["maxWorkers"] = maxWorkers
+            struct["maxWorkers"] = maxWorkers
         if seekTo:
-            self._struct["attributes"]["seekTo"] = seekTo
+            struct["attributes"]["seekTo"] = seekTo
         if readBatchSize:
-            self._struct["attributes"]["readBatchSize"] = readBatchSize
+            struct["attributes"]["readBatchSize"] = readBatchSize
         if partitions:
-            self._struct["attributes"]["partitions"] = partitions
+            struct["attributes"]["partitions"] = partitions
         if pollingIntervalMS:
-            self._struct["attributes"]["pollingIntervalMs"] = pollingIntervalMS
-        access_key = access_key if access_key else environ["V3IO_ACCESS_KEY"]
-        self._struct["password"] = access_key
+            struct["attributes"]["pollingIntervalMs"] = pollingIntervalMS
+        access_key = access_key \
+            if access_key else environ.get("V3IO_ACCESS_KEY")
+        if not access_key:
+            raise ValueError('access_key must be set '
+                             '(via argument or environ V3IO_ACCESS_KEY)')
+        struct["password"] = access_key
+
+        super(V3IOStreamTrigger, self).__init__(struct)
+
+    @property
+    def get_url(self):
+        return self._struct["url"]
